/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: ColorList.C,v $
 *	$Author: dalke $	$Locker:  $		$State: Exp $
 *	$Revision: 1.22 $	$Date: 1997/02/07 21:03:55 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * A collection of colored materials; provides ways to define them and edit
 * them.  Each Displayable must turn on or off material characteristics
 * if needed, and say which color to use.
 *
 ***************************************************************************/

#include <stdio.h>
#include "ColorList.h"
#include "Displayable.h"
#include "DispCmds.h"
#include "DisplayDevice.h"
#include "NameList.h"
#include "Inform.h"

// globally-accessible strings with methods for creating color scale
char *colorScaleMethod[ColorList::SCALE_TOTAL] =
{ "RGB", "BGR", "RWB", "BWR", 
  "RWG", "GWR", "GWB", "BWG",
  "BlkW", "WBlk"};

// globally-accessible strings with names for the regular colors
char *defColorNames[REGCLRS] = { "blue", "red", "gray", "orange", "yellow",
	"tan", "silver", "green", "white", "pink", "cyan", "purple",
	"lime", "mauve", "ochre", "iceblue", "black" };

// static data for default colors (same as diffuse characteristic)
static float defaultColor[REGCLRS][3] = {
        { 0.25, 0.25, 1.00 }, { 1.00, 0.0, 0.0 }, // BLUE, RED
        { 0.35, 0.35, 0.35 }, { .8, 0.5, .2 },  // GREY, ORANGE
        { 0.8, 0.8, 0.0 }, { 0.5, 0.5, 0.2 },   // YELLOW, TAN
        { 0.6, 0.6, 0.6 }, { 0.2, 0.7, 0.2 },   // SILVER, GREEN
        { 1.0, 1.0, 1.0 }, { 1.0, 0.6, 0.6 },   // WHITE, PINK
        { 0.25, 0.75, 0.75 }, { 0.65, 0.3, 0.65 },        // CYAN, PURPLE
        { 0.5, 0.9, 0.4 }, { 0.9, 0.4, 0.7 },   // LIME, MAUVE
        { 0.5, 0.3, 0.0 }, { 0.5, 0.75, 0.75 }, // OCHRE, ICEBLUE
	{ 0.0, 0.0, 0.0 }			// BLACK
};

static float defaultSpecular[3] = { 1.0, 1.0, 1.0 };
static float defaultAmbient[3] = { 0.0, 0.0, 0.0 };
static float defaultShininess = 40.0;
static float defaultAlphaSolid = 1.0;
static float defaultAlphaTrans = 0.3;


// constructor
ColorList::ColorList(Scene *sc) : Displayable3D(MULT, "Color List", sc, 5) {
  int i, j;
  float *rcol, *tcol;
  char cNameNumBuf[20];
  
  dispCmdColorDef = new DispCmdColorDef;

  // initialize flags
  for(i=0; i < MAXCOLORS; needDefine[i++] = TRUE);
  needAnyDefined = TRUE;
  needDefineCleanup = FALSE;
  
  // initialize color data for 'regular' colors
  for(i=0; i < REGCLRS; i++) {
  
    // add color names to colorNames list
    colorNames.add_name(defColorNames[i], i);

    // initialize color data
    reset_color(i);
  }

  // add transparent color names
  for(i=0; i < REGCLRS; i++) {
    sprintf(cNameNumBuf, "trans_%s", defColorNames[i]);
    colorNames.add_name(cNameNumBuf, i + REGCLRS);
  }

  // initialize some color data for 'colormap' colors
  for(i=0; i < MAPCLRS; i++) {
    rcol = colorData[BEGMAP + i];
    tcol = colorData[BEGTRANMAP + i];
    for(j=0; j < 3; j++) {
      rcol[AMBIENT_INDEX + j]  = tcol[AMBIENT_INDEX + j] = defaultAmbient[j];
      rcol[SPECULAR_INDEX + j] = tcol[SPECULAR_INDEX + j]= defaultSpecular[j];
    }

    // assign values to other components
    rcol[SHININESS_INDEX] 	= tcol[SHININESS_INDEX] = defaultShininess;
    rcol[ALPHA_INDEX]		= defaultAlphaSolid;
    tcol[ALPHA_INDEX]		= defaultAlphaTrans;
  }
  
  // add the color map names
  for (i=0; i<MAPCLRS*2; i++) {
    sprintf(cNameNumBuf, "%d", i+2*REGCLRS);
    colorNames.add_name(cNameNumBuf, i+2*REGCLRS);
  }

  // initialize data for rest of color scale
  scaleMethod = RGB;
  scaleMin = 0.1;
  scaleMax = 1.0;
  scaleMidpoint = 0.5;
  create_colorscale();

  // Displayable characteristics
  rot_off();
  scale_off();
  glob_trans_off();
  cent_trans_off();
  
  MSGDEBUG(1,"Creating Color List with " << MAXCOLORS << " colors ...");
  MSGDEBUG(2,sendmsg);
}


// destructor
ColorList::~ColorList(void) {
  // delete all added color categories
  for(int i=0; i < categories.num(); i++) {
    delete (categories.data(i));
    categories.set_data(i,(NameList<int>*) NULL);
  }

  delete dispCmdColorDef;
}


// create the color scale based on current settings

////////////////
// definitions for scales with 3 colors
// 1 at val = 0; 0 at mid and greater
static float slope3_lower(float mid, float val) {
  if (val > mid) return 0.0;
  if (mid == 0) return 1.0;
  return (1.0 - val / mid);
}

// 1 at val = 1; 0 at mid and lesser
static float slope3_upper(float mid, float val) {
  if (val < mid) return 0.0;
  if (mid == 1) return 1.0;
  return (1.0 - (1.0 - val) / (1.0 - mid));
}

// 1 at val = mid; 0 at val = 0 an 1
static float slope3_middle(float mid, float val) {
  if (val < mid) {
    return val / mid;
  }
  if (val > mid) {
    return (1-val) / (1-mid);
  }
  return 1.0;
}


// given RGB values at (0,mid,1), return the correct RGB triplet
static void scale_color(float *lowRGB, float *midRGB, float *highRGB,
			float mid, float val, float scaleMin, float *rgb)
{
  float w1, w2, w3;
  rgb[0] =
    (lowRGB[0]  * (w1 = slope3_lower(mid, val)) +
     midRGB[0]  * (w2 = slope3_middle(mid, val)) +
     highRGB[0] * (w3 = slope3_upper(mid, val))) /
    (w1 + w2 + w3) +
    scaleMin;
  if (rgb[0] < 0) rgb[0] = 0; else if (rgb[0] > 1.0) rgb[0] = 1.0;

  rgb[1] =
    (lowRGB[1]  * w1 +
     midRGB[1]  * w2 +
     highRGB[1] * w3) /
    (w1 + w2 + w3) +
    scaleMin;
  if (rgb[1] < 0) rgb[1] = 0; else if (rgb[1] > 1.0) rgb[1] = 1.0;

  rgb[2] =
    (lowRGB[2]  * w1 +
     midRGB[2]  * w2 +
     highRGB[2] * w3) /
    (w1 + w2 + w3) +
    scaleMin;
  if (rgb[2] < 0) rgb[2] = 0; else if (rgb[2] > 1.0) rgb[2] = 1.0;
}


void ColorList::create_colorscale(void) {
  // make sure settings and other values are OK
  if(scaleMin < -1.0) {
    scaleMin = -1.0;
  } else if (scaleMin > 1.0) {
    scaleMin = 1.0;
  }
  if(scaleMidpoint < 0 || scaleMidpoint > 1.0)
    scaleMidpoint = 0.5;

  float *rcol, *tcol;

  float low[3];
  float mid[3];
  float high[3];

  switch (scaleMethod) {
  default:
  case RGB:
    low[0]  = 1.0; low[1]  = 0.0; low[2]  = 0.0;
    mid[0]  = 0.0; mid[1]  = 1.0; mid[2]  = 0.0;
    high[0] = 0.0; high[1] = 0.0; high[2] = 1.0;
    break;

  case BGR:
    low[0]  = 0.0; low[1]  = 0.0; low[2]  = 1.0;
    mid[0]  = 0.0; mid[1]  = 1.0; mid[2]  = 0.0;
    high[0] = 1.0; high[1] = 0.0; high[2] = 0.0;
    break;

  case RWB:
    low[0]  = 1.0; low[1]  = 0.0; low[2]  = 0.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 0.0; high[1] = 0.0; high[2] = 10;
    break;

  case BWR:
    low[0]  = 0.0; low[1]  = 0.0; low[2]  = 1.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 1.0; high[1] = 0.0; high[2] = 0.0;
    break;

  case RWG:
    low[0]  = 1.0; low[1]  = 0.0; low[2]  = 0.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 0.0; high[1] = 1.0; high[2] = 0.0;
    break;

  case GWR:
    low[0]  = 0.0; low[1]  = 1.0; low[2]  = 0.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 1.0; high[1] = 0.0; high[2] = 0.0;
    break;

  case GWB:
    low[0]  = 0.0; low[1]  = 1.0; low[2]  = 0.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 0.0; high[1] = 0.0; high[2] = 1.0;
    break;

  case BWG:
    low[0]  = 0.0; low[1]  = 0.0; low[2]  = 1.0;
    mid[0]  = 1.0; mid[1]  = 1.0; mid[2]  = 1.0;
    high[0] = 0.0; high[1] = 1.0; high[2] = 0.0;
    break;

  case BLK_W:
    low[0]  = 0.0; low[1]  = 0.0; low[2]  = 0.0;
    mid[0]  = 0.5; mid[1]  = 0.5; mid[2]  = 0.5;
    high[0] = 1.0; high[1] = 1.0; high[2] = 1.0;
    break;

  case W_BLK:
    low[0]  = 1.0; low[1]  = 1.0; low[2]  = 1.0;
    mid[0]  = 0.5; mid[1]  = 0.5; mid[2]  = 0.5;
    high[0] = 0.0; high[1] = 0.0; high[2] = 0.0;
    break;
  }
  

  for (int i=0; i<MAPCLRS; i++) {
    rcol = colorData[BEGMAP + i];
    tcol = colorData[BEGTRANMAP + i];
    
    // define the diffuse color
    float relpos = float(i) / float(MAPCLRS - 1);
    
    // get the r, g, and b values
    scale_color(low, mid, high, scaleMidpoint, relpos, scaleMin,
		rcol+DIFFUSE_INDEX);
    tcol[DIFFUSE_INDEX + 0] = rcol[DIFFUSE_INDEX + 0]; // copy into the
    tcol[DIFFUSE_INDEX + 1] = rcol[DIFFUSE_INDEX + 1]; // translucent
    tcol[DIFFUSE_INDEX + 2] = rcol[DIFFUSE_INDEX + 2]; // color
  }


#ifdef OLD_COLOR_SCALE
  // find 'middle' color, and sizerange of colors
  float range = scaleMax - scaleMin;
  float midcolor = 0.5 * (scaleMax + scaleMin);

  // precalc the 'quarterpoints'
  float leftmidpt = 0.5 * scaleMidpoint;
  float rightmidpt = 0.5 * (scaleMidpoint + 1.0);

  // precalc some slopes for graphs
  float pslope, nslope, leftslope, rightslope;
  if(scaleMidpoint > 0.0)
    nslope = - (range / scaleMidpoint);
  if(scaleMidpoint < scaleMax)
    pslope = range / (1.0 - scaleMidpoint);
  if(leftmidpt < scaleMidpoint)
    leftslope = range / (scaleMidpoint - leftmidpt);
  if(rightmidpt > scaleMidpoint)
    rightslope = range / (scaleMidpoint - rightmidpt);

  // create the color scale
  // initialize color data for 'colormap' colors
  for(int i=0; i < MAPCLRS; i++) {
    rcol = colorData[BEGMAP + i];
    tcol = colorData[BEGTRANMAP + i];
    
    // define the diffuse color, which determines the color of the material
    // color map goes from blue (small numbers) through green (mid numbers)
    // to red (large numbers)
    float relpos = (float)i / (float)(MAPCLRS - 1);
    float rcolor, gcolor, bcolor;

    if(scaleMethod == RGB) {
      rcolor = gcolor = bcolor = scaleMin;

      // set red component
      if(relpos > scaleMidpoint) 
        rcolor += pslope * (relpos - scaleMidpoint);

      // set green component
      if(relpos > leftmidpt && relpos <= scaleMidpoint) 
        gcolor += leftslope * (relpos - leftmidpt);
      else if(relpos >= scaleMidpoint && relpos < rightmidpt)
        gcolor += range + rightslope * (relpos - scaleMidpoint);

      // set blue component
      if(relpos < scaleMidpoint) 
        bcolor += range + (nslope * relpos);

    } else if(scaleMethod == RWB) {
      rcolor = gcolor = bcolor = midcolor;

      // set red component
      if(relpos < scaleMidpoint)
        rcolor += range + (0.5 * nslope * relpos); 

      // set blue component (green is always midcolor)
      if(relpos > scaleMidpoint)
        bcolor += 0.5 * pslope * (relpos - scaleMidpoint);
   } else if (scaleMethod == BLK_W) {
      if(relpos < scaleMidpoint)  // slope up to half
        rcolor = gcolor = bcolor = 0 + (relpos - 0.0) /
                   (scaleMidpoint - 0.0)*0.5;

      if(relpos >= scaleMidpoint)  // from half to full
        rcolor = gcolor = bcolor = 0.5 + (relpos - scaleMidpoint) /
	           (1 - scaleMidpoint) * 0.5;
   }

    // assign values to components
    rcol[DIFFUSE_INDEX + 0]	= tcol[DIFFUSE_INDEX + 0] = rcolor;
    rcol[DIFFUSE_INDEX + 1]	= tcol[DIFFUSE_INDEX + 1] = gcolor;
    rcol[DIFFUSE_INDEX + 2]	= tcol[DIFFUSE_INDEX + 2] = bcolor;    
  }
#endif

  for(int csn=0; csn < MAPCLRS; csn++)
    needDefine[MAPCOLOR(csn)] = needDefine[MAPTRANSLUCENT(csn)] = TRUE;
  // signal we need these colors redefined
  needAnyDefined = TRUE;
}


// add new color category; return it's index (may already have been added)
int ColorList::add_color_category(char *newcat) {
  int oldnum, retval;

  if(!newcat)
    return (-1);
 
  // add new name with NULL category; see what new index is
  oldnum = categories.num();
  retval = categories.add_name(newcat,(NameList<int>*) NULL);
  if(oldnum != categories.num()) {
    // number of categories changed, so we must have added a new one.
    // create a new NameList, and put it in the data value for the new cat.
    categories.set_data(retval, new NameList<int>);

    MSGDEBUG(2,"Added new color category '" << newcat << "', index="<< retval);
    MSGDEBUG(2,sendmsg);
  } else {
    MSGDEBUG(2,"Existing color category '" << newcat << "', index=" << retval);
    MSGDEBUG(2," found." << sendmsg);
  }
  
  return retval;
}


//
// change data for the Nth color
//
  
void ColorList::change_alpha(int n, float a) {
  if(n >= 0 && n < MAXCOLORS) {
    colorData[n][ALPHA_INDEX] = a;
    needDefine[n] = needAnyDefined = TRUE;
  }
}

void ColorList::change_shininess(int n, float s) {
  if(n >= 0 && n < MAXCOLORS) {
    colorData[n][SHININESS_INDEX] = s;
    needDefine[n] = needAnyDefined = TRUE;
  }
}

void ColorList::change_ambient(int n, float r, float g, float b) {
  if(n >= 0 && n < MAXCOLORS) {
    colorData[n][AMBIENT_INDEX] = r;
    colorData[n][AMBIENT_INDEX + 1] = g;
    colorData[n][AMBIENT_INDEX + 2] = b;
    needDefine[n] = needAnyDefined = TRUE;
  }
}

void ColorList::change_specular(int n, float r, float g, float b) {
  if(n >= 0 && n < MAXCOLORS) {
    colorData[n][SPECULAR_INDEX] = r;
    colorData[n][SPECULAR_INDEX + 1] = g;
    colorData[n][SPECULAR_INDEX + 2] = b;
    needDefine[n] = needAnyDefined = TRUE;
  }
}

void ColorList::change_diffuse(int n, float r, float g, float b) {
  if(n >= 0 && n < MAXCOLORS) {
    colorData[n][DIFFUSE_INDEX] = r;
    colorData[n][DIFFUSE_INDEX + 1] = g;
    colorData[n][DIFFUSE_INDEX + 2] = b;
    needDefine[n] = needAnyDefined = TRUE;
  }
}

void ColorList::change_color(int n, float r, float g, float b) {
  change_diffuse(n, r, g, b);
}


// restore the Nth color to it's original form
void ColorList::reset_color(int n) {
  if(n >= BEGREGCLRS && n < (BEGREGCLRS + REGCLRS)) {
    float *rcol = colorData[BEGREGCLRS + n];
    float *tcol = colorData[BEGTRAN + n];
    for(int j=0; j < 3; j++) {
      rcol[AMBIENT_INDEX + j]  = tcol[AMBIENT_INDEX + j] = defaultAmbient[j];
      rcol[SPECULAR_INDEX + j] = tcol[SPECULAR_INDEX + j]= defaultSpecular[j];
      rcol[DIFFUSE_INDEX + j]  = tcol[DIFFUSE_INDEX + j] = defaultColor[n][j];
    }
    rcol[SHININESS_INDEX] = tcol[SHININESS_INDEX] = defaultShininess;
    rcol[ALPHA_INDEX] = defaultAlphaSolid;
    tcol[ALPHA_INDEX] = defaultAlphaTrans;
    
    needDefine[n] = needAnyDefined = needDefine[n+REGCLRS] = TRUE;
  }
}

// given (r,g,b) find the color index in the current list which
// best matches it in the least square sense
int ColorList::nearest_index(float r, float g, float b)
{
   float *rcol = colorData[BEGREGCLRS];  // get the solid colors
   float lsq = r - rcol[DIFFUSE_INDEX]; lsq *= lsq;
   float tmp = g - rcol[DIFFUSE_INDEX+1]; lsq += tmp * tmp;
         tmp = b - rcol[DIFFUSE_INDEX+1]; lsq += tmp * tmp;
   float best = lsq;
   int bestidx = BEGREGCLRS;
   for (int n= BEGREGCLRS+1; n < (BEGREGCLRS + REGCLRS); n++) {
      rcol = colorData[BEGREGCLRS+n];
      lsq = r - rcol[DIFFUSE_INDEX]; lsq *= lsq;
      tmp = g - rcol[DIFFUSE_INDEX+1]; lsq += tmp * tmp;
      tmp = b - rcol[DIFFUSE_INDEX+2]; lsq += tmp * tmp;
      if (lsq < best) {
       best = lsq;
       bestidx = n;
      }
   }
   return bestidx;
}


// create the command list
void ColorList::create_cmdlist(void) {
  // do reset first
  reset_disp_list();
  
  // check if need to define any materials
  if(needAnyDefined) {
    for(int i=0; i < MAXCOLORS; i++) {
      if(needDefine[i]) {
        dispCmdColorDef->putdata(i, colorData[i], this);
	needDefine[i] = FALSE;
      }
    }
  }

  // reset flags
  needDefineCleanup = needAnyDefined;
  needAnyDefined = FALSE;
}


// prepare the colorlist for display ... regenerate command list if necessary
void ColorList::prepare(DisplayDevice *) {
  // if a flag has been set, change the settings
  if(needAnyDefined || needDefineCleanup)
    create_cmdlist();
}

