/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: DisplayDevice.C,v $
 *	$Author: dalke $	$Locker:  $		$State: Exp $
 *	$Revision: 1.26 $	$Date: 1997/03/24 22:15:06 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * DisplayDevice - abstract base class for all particular objects which
 *	can process a list of drawing commands and render the drawing
 *	to some device (screen, file, preprocessing script, etc.)
 *
 ***************************************************************************/

#include <math.h>
#include "MouseEvent.h"
#include "DisplayDevice.h"
#include "Inform.h"
#include "DispCmds.h"
#include "ColorList.h"
#include "utilities.h"
#include "Mouse.h"     // for WAIT_CURSOR

// static data for this object
static char *stereoNameStr[1] = { "Stereo Off" };

static char *detail_level_names[] = {
  "Nothing", "Points", "Wireframe", "Flat", "Full" };

/////////////////////////  constructor and destructor  
DisplayDevice::DisplayDevice
(char *nm) : transMat(16) {
  int i;

  // save name
  name = stringdup(nm);

  MSGDEBUG(1,"Creating new DisplayDevice '" << name << "'." << sendmsg);

  // set drawing characteristics default values
  lineStyle = ::SOLIDLINE;
  lineWidth = 1;
  sphereRes = 3;
  sphereMode = ::SOLIDSPHERE;
  backColor[0] = backColor[1] = backColor[2] = 0.0;
  my_detail_level = max_detail_level();
  my_alt_detail_level = my_detail_level;
  my_depth_sort = FALSE;

  // set scalar values
  has2D = TRUE;
  Dim = 3;
  aaAvailable = cueingAvailable = FALSE;
  aaEnabled = cueingEnabled = FALSE;
  xOrig = yOrig = xSize = ySize = 0;
  pickRegion = 0.01;
  colorCat = (-1);

  // set viewing geometry ... looking from z-axis in negative direction,
  // with 90-degree field of view and assuming the origin is in the
  // center of the viewer's 'screen'
  nearClip = 0.5;
  farClip = 10.0;
  eyePos[0] = eyePos[1] = 0.0;  eyePos[2] = 2.0;
  set_screen_pos(2.0 * eyePos[2], 0.0, 4.0/3.0);

  // set stereo options ... while there is no stereo mode by default,
  // set up normal values for stereo data
  inStereo = 0;
  stereoModes = 1;
  stereoNames = stereoNameStr;
  eyeSep = 0.065;               // default eye seperation
  eyeDist = eyePos[2];		// assumes viewer on pos z-axis
  eyeDir[0] = -eyePos[0];  eyeDir[1] = -eyePos[1];  eyeDir[2] = -eyePos[2];
  upDir[0] = upDir[2] = 0.0;  upDir[1] = 1.0;
  calc_eyedir();		// determines eye separation direction

  my_projection = PERSPECTIVE;

  // set internal flags for lights
  for(i=0; i < DISP_LIGHTS; i++)
    lightDefined[i] = lightOn[i] = FALSE;
  for(i=0; i < MAXCOLORS; i++)
    matDefined[i] = FALSE;
  materialOn = (-1);
  materialsActive = TRUE;
    
  // load identity matrix onto top of transformation matrix stack
  Matrix4 temp_ident;
  transMat.push(temp_ident);

  for (i = 0; i < MouseEvent::B_NUM_BUTTONS; i++)
    buttonDown[i] = FALSE;
  mouseX = mouseY = 0;
}


// destructor
DisplayDevice::~DisplayDevice(void) {

  set_stereo_mode(0);		// return to non-stereo, if necessary
  delete [] name;
}

///////////////////////// return the detail names
const char *DisplayDevice::detail_name(int i) {
  if (i < MIN_DETAIL_LEVEL) i = MIN_DETAIL_LEVEL;
  if (i > MAX_DETAIL_LEVEL) i = MAX_DETAIL_LEVEL;
  return detail_level_names[i];
}



/////////////////////////  protected nonvirtual routines  
// calculate the position of the near frustrum plane, based on current values
// of Aspect, vSize, zDist, nearClip and eyePosition
// Assumes the viewer is looking toward the xy-plane
void DisplayDevice::calc_frustrum(void) {
  float d, halfvsize = 0.5 * vSize;
  float halfhsize = Aspect * halfvsize;

  if(eyePos[2] - zDist != 0.0) {
    d = nearClip / (eyePos[2] - zDist);
    cpRight = d * halfhsize;
    cpLeft = -cpRight;
    cpUp = d * halfvsize;
    cpDown = -cpUp;
  }
}


// calculate eyeSepDir, based on up vector and look vector
// eyeSepDir = 1/2 * eyeSep * (lookdir x updir) / mag(lookdir x updir)
void DisplayDevice::calc_eyedir(void) {
  float *L = eyeDir;
  float *U = upDir;
  float m, A = 0.5 * eyeSep;
  eyeSepDir[0] = L[1] * U[2] - L[2] * U[1];
  eyeSepDir[1] = L[2] * U[0] - L[0] * U[2];
  eyeSepDir[2] = L[0] * U[1] - L[1] * U[0];
  m = sqrtf(eyeSepDir[0] * eyeSepDir[0] + eyeSepDir[1] * eyeSepDir[1] +
            eyeSepDir[2] * eyeSepDir[2]);
  if(m > 0.0)
    A /= m;
  else
    A = 0.0;
  eyeSepDir[0] *= A;
  eyeSepDir[1] *= A;
  eyeSepDir[2] *= A;
}

/////////////////////////  protected virtual routines (default values)
// virtual routines to deal with light sources and materials at device level
// these just return TRUE, but do nothing
int DisplayDevice::do_define_light(int, float *, float *) {
  return TRUE;
}

int DisplayDevice::do_activate_light(int, int) {
  return TRUE;
}

int DisplayDevice::do_define_material(int, float *) {
  return TRUE;
}

int DisplayDevice::do_activate_material(int, int) {
  return TRUE;
}

// virtual function which is called when a new colorlist is provided.
// This should NOT have to be changed by derived classes.
void DisplayDevice::do_use_colors(void) {

  // create a display color category ... if it already exists, we'll get
  // it here
  colorCat = colorList->add_color_category("Display");
  
  // add colors to the category, if they do not yet exist
  if((colorList->color_category(colorCat))->typecode("Background") < 0)
    (colorList->color_category(colorCat))->add_name("Background", REGBLACK);

  // update the colors by indicating it has changed
  do_color_changed(colorList, colorCat);
}

// do action due to the fact that a color for the given ColorList for
// the specified category has changed
// This should NOT have to be changed by derived classes.
void DisplayDevice::do_color_changed(ColorList *collist, int ccat) {
  if(collist == colorList && ccat == ccat) {
    // must be changing the background, since this is the only color here
    float *bgcolor = colorList->color(
    	(colorList->color_category(colorCat))->data("Background"));
    set_background(bgcolor[0], bgcolor[1], bgcolor[2]);
  }
}


/////////////////////////  public nonvirtual routines  

// Copy all relevant properties from one DisplayDevice to another
DisplayDevice& DisplayDevice::operator=( DisplayDevice &display) {
  int i,j;
  
  colorCat = display.colorCat;
  
  xOrig = display.xOrig;
  yOrig = display.yOrig;
  xSize = display.xSize;
  ySize = display.ySize;
  
  // Do something about the stack.  For the moment, only copy the top
  // item on the stack.
  if(transMat.num() > 0) {
    transMat.pop();
  }
  transMat.push( (display.transMat).top() );
  
  for(i=0; i<3; i++) {
      eyePos[i] = display.eyePos[i];
      eyeDir[i] = display.eyeDir[i];
      upDir[i] = display.upDir[i];
      eyeSepDir[i] = display.eyeSepDir[i];
      backColor[i] = display.backColor[i];
  }
  
  nearClip = display.nearClip;
  farClip = display.farClip;
  vSize = display.vSize;
  zDist = display.zDist;
  Aspect = display.Aspect;
  cpUp = display.cpUp;
  cpDown = display.cpDown;
  cpLeft = display.cpLeft;
  cpRight = display.cpRight;
  inStereo = display.inStereo;
  eyeSep = display.eyeSep;
  eyeDist = display.eyeDist;
  lineStyle = display.lineStyle;
  lineWidth = display.lineWidth;
  my_detail_level = display.my_detail_level;
  my_alt_detail_level = display.my_alt_detail_level;
  my_depth_sort = display.my_depth_sort;
  
  for(i=0;i<DISP_LIGHTS;i++) {
      lightDefined[i] = display.lightDefined[i];
      lightOn[i] = display.lightOn[i];
      for(j=0;j<3;j++) {
          lightColor[i][j] = display.lightColor[i][j];
          lightPos[i][j] = display.lightPos[i][j];
      }
  }
  
  for(i=0;i<MAXCOLORS;i++) {
      matDefined[i] = display.matDefined[i];
      for(j=0;j<COLOR_ITEMS;j++) {
          matData[i][j] = display.matData[i][j];
      }
  }
  
  materialOn = display.materialOn;
  materialsActive = display.materialsActive;
  
  return *this;
}


// return relative 2D screen coordinates, given 2D or 3D world coordinates.
void DisplayDevice::rel_screen_loc_3D(float *wloc, float *sloc) {
  // first, find absolute screen coordinates
  long abscoords[2];
  abs_screen_loc_3D(wloc, abscoords);
  
  // now convert these to relative coordss
  sloc[0] = (float) (abscoords[0]);
  sloc[1] = (float) (abscoords[1]);
  rel_screen_pos(sloc[0], sloc[1]);
}

void DisplayDevice::rel_screen_loc_2D(float *wloc, float *sloc) {
  // first, find absolute screen coordinates
  long abscoords[2];
  abs_screen_loc_2D(wloc, abscoords);
  
  // now convert these to relative coordss
  sloc[0] = (float) (abscoords[0]);
  sloc[1] = (float) (abscoords[1]);
  rel_screen_pos(sloc[0], sloc[1]);
}


//
// routines to deal with light sources
//
  
void DisplayDevice::define_light(int n, float *color, float *position) {
  int i;
  
  if(n < 0 || n >= DISP_LIGHTS)
    return;
  
  // save data for future reference
  for(i=0; i < 3; i++) {
    lightColor[n][i] = color[i];
    lightPos[n][i] = position[i];
  }
  
  // have derived class actually do the action
  lightDefined[n] = do_define_light(n, color, position);
}

void DisplayDevice::activate_light(int n, int turnon) {
  if(n < 0 || n >= DISP_LIGHTS || !lightDefined[n])
    return;
    
  if(do_activate_light(n, turnon))
    lightOn[n] = turnon;
}


//
// routines to deal with material definitions
//

void DisplayDevice::define_material(int n, float *data) {

  if(n < 0 || n >= MAXCOLORS)
    return;
    
  // save data for future reference
  for(int i=0; i < COLOR_ITEMS; i++)
    matData[n][i] = data[i];

  // have derived class actually do the action
  matDefined[n] = do_define_material(n, data);
}

// turn on the specified material
void DisplayDevice::activate_material(int n) {

  if(n < 0 || n >= MAXCOLORS)
    return;

  if(do_activate_material(n, TRUE)) {
    materialOn = n;
    materialsActive = TRUE;
  }
}

// turn on the current material (if any)
void DisplayDevice::activate_material(void) {
  int n = materialOn;

  if(n < 0 || n >= MAXCOLORS || !matDefined[n])
    return;

  if(do_activate_material(n, TRUE))
    materialsActive = TRUE;
}

// turn off the use of materials
void DisplayDevice::deactivate_materials(void) {

  if(!materialsActive)
    return;
    
  if(do_activate_material(materialOn, FALSE))
    materialsActive = FALSE;
}

/////////////////////////  public virtual routines

//
// event handling routines
//

// queue the standard events (need only be called once ... but this is
// not done automatically by the window because it may not be necessary or
// even wanted)
void DisplayDevice::queue_events(void) { return; }

// test if there is an event ready
int DisplayDevice::test_events(void) { return FALSE; }

// read the next event ... returns an event type (one of the above ones),
// and a value.  Returns success, and sets arguments.
int DisplayDevice::read_event(long &, long &) { return WIN_NOEVENT; }

// set the mouse state, for non-polling interfaces
void DisplayDevice::record_mouse_event(const MouseEvent &ev) {
  if (ev.type() == MouseEvent::BUTTON_DOWN ||
      ev.type() == MouseEvent::BUTTON_UP) {
    if (ev.button() >= 0 && ev.button() < MouseEvent::B_NUM_BUTTONS)
      buttonDown[ev.button()] = (ev.type() == MouseEvent::BUTTON_DOWN);
  }
  mouseX = ev.x();
  mouseY = ev.y();
}

//
// get the current state of the device's pointer (i.e. cursor if it has one)
//

// abs pos of cursor from lower-left corner
int DisplayDevice::x(void) { return mouseX; }

// same, for y direction
int DisplayDevice::y(void) { return mouseY; }

// whether a button is currently pressed
int DisplayDevice::button_down(int b) { return buttonDown[b]; }

// the shift state (shift key, control key, and/or alt key)
int DisplayDevice::shift_state(void) {
  return 0; // by default, nothing is down
}

// allow the user to define a special shape for the cursor ... the data
// is assumed to consist of a single-dim array of 32 unsigned shorts; the
// first 16 define a 16x16-bit (row-major) pattern, the last 16 are a
// "second layer" which may be drawn in a different color
// args: which shape (>0, with 0 the "default" case which cannot be changed)
// the bitmap data, and the "hot spot" from the lower-left corner
void DisplayDevice::change_cursor_shape(int, unsigned short *, int, int) { }

// set the Nth cursor shape as the current one.  If no arg given, the
// default shape (n=0) is used.
void DisplayDevice::set_cursor(int) { }

int DisplayDevice::get_cursor(void) { return 0;}

//  put the cursor into the WAIT cursor
void DisplayDevice::start_wait(void) {
  old_cursor = get_cursor();
  set_cursor(WAIT_CURSOR);
}
//  return the cursor to the old state
void DisplayDevice::stop_wait(void) {
  set_cursor(old_cursor);
}

//
// virtual routines to create and activate a popup (possibly pull-down) menu 
//

// given a PopupMenu definition, create it in the window (but do not
// activate it yet).  Return success.
int DisplayDevice::menu_create(PopupMenu *) { return TRUE; }

// activate a previously-created menu.  If the windowing system has no
// current menu, or cannot do this operation, returns (-1); otherwise,
// returns the return code for the select (-1 if none selected).
int DisplayDevice::menu_activate(void) { return -1; }

// delete the given menu from the display.
// If no argument is given, deletes the current menu (if any).
// Returns success.
int DisplayDevice::menu_delete(PopupMenu *) { return TRUE; }

// return the eye position to it's original position
void DisplayDevice::reset_eye(void) {
  move_eye(0.0, 0.0, 2.0);
  change_up(0.0, 1.0, 0.0);
  change_look(0.0, 0.0, - eye_dist());
}


// virtual functions to turn on/off depth cuing and antialiasing
void DisplayDevice::aa_on(void) { }
void DisplayDevice::aa_off(void) { }
void DisplayDevice::cueing_on(void) { }
void DisplayDevice::cueing_off(void) { }


// return absolute 2D screen coordinates, given 2D or 3D world coordinates.
void DisplayDevice::abs_screen_loc_3D(float *wloc, long *sloc) {
  // just return world coords
  for(int i=0; i < 2; i++)
    sloc[i] = (long)(wloc[i]);
}

void DisplayDevice::abs_screen_loc_2D(float *wloc, long *sloc) {
  // just return world coords
  for(int i=0; i < 2; i++)
    sloc[i] = (long)(wloc[i]);
}

// size of physical display object
void DisplayDevice::screen_size_mm(long& x, long& y) {
  x = y = 1000;		// just say it's a meter
}

void DisplayDevice::screen_size_pixel(long& x, long& y) {
  x = y = 1000;		// just say 1000 pixels
}


// Given a 3D point (pos A),
// and a 2D rel screen pos point (for pos B), computes the 3D point
// which goes with the second 2D point at pos B.  Result returned in 3rd arg.
void DisplayDevice::find_3D_from_2D(float *, float *, float *) { }
/*
// Given a 3D point (pos A),
// and a 2D rel screen pos point (for pos B), computes the 3D point
// which goes with the second 2D point at pos B.  Result returned in B3D.
void DisplayDevice::find_3D_from_2D(float *A2D, float *A3D,
					float *B2D, float *B3D) {
  float tmpB2D[2], trans2D[3], trans2Dtmp[3], trans3Dtmp[3];
  
  // Add trans to A3D in X&Y dir equal to difference between given 2D points
  trans2D[0] = B2D[0] - A2D[0];
  trans2D[1] = B2D[1] - A2D[1];
  trans2D[2] = 0.0;
  add(B3D, A3D, trans2D);
  
  // now convert the translated 3D point to rel screen coords
  rel_screen_loc_3D(B3D, tmpB2D);
  
  // compute difference between new 2D point and starting point
  trans2Dtmp[0] = tmpB2D[0] - A2D[0];
  trans2Dtmp[1] = tmpB2D[1] - A2D[1];
  trans2Dtmp[2] = 0.0;

  // from ratio of 2D differences, times change in 3D tmp difference, get
  // change in 3D for wanted coordinates
  subtract(trans3Dtmp, B3D, A3D);
  B3D[0] = (trans2Dtmp[0] != 0.0 ?
  	A3D[0] + (trans3Dtmp[0] / trans2Dtmp[0]) * trans2D[0] :  A3D[0]);
  B3D[1] = (trans2Dtmp[1] != 0.0 ?
  	A3D[1] + (trans3Dtmp[1] / trans2Dtmp[1]) * trans2D[1] :  A3D[1]);
  B3D[2] = A3D[2];
}
*/




// change to a different stereo mode (0 means 'off')
void DisplayDevice::set_stereo_mode(int sm) {
  if(sm != 0) {
    msgErr << "DisplayDevice: Illegal stereo mode " << sm << " specified."
           << sendmsg;
  } else {
    inStereo = sm;
  }
}

// save the current transformation matrix ... by default, a copy is
// kept in this object, but derived classes may handle the trans matrix
// differently
void DisplayDevice::push(void) {
  transMat.push();
}

// restore the current transformation matrix
void DisplayDevice::pop(void) {
  transMat.pop();
}

// replace the current trans matrix with the given one
void DisplayDevice::loadmatrix(Matrix4 &m) {
  (transMat.top()).loadmatrix(m);
}

// multiply the current trans matrix with the given one
void DisplayDevice::multmatrix(Matrix4 &m) {
  (transMat.top()).multmatrix(m);
}


//
// virtual routines for preparing to draw, drawing, and finishing drawing
//
  
int DisplayDevice::prepare2D(int) { return 1;}	// ready to draw 2D
int DisplayDevice::prepare3D(int) { return 1;}	// ready to draw 3D
void DisplayDevice::clear(void) { }		// erase the device
void DisplayDevice::left(void) { }		// ready to draw left eye
void DisplayDevice::right(void) { }		// ready to draw right eye
void DisplayDevice::normal(void) { }		// ready to draw non-stereo
void DisplayDevice::update(int) { }		// finish up after drawing
void DisplayDevice::reshape(void) { }		// refresh device after change
  

//
//*******************  the rendering routine  *********************
//
void DisplayDevice::render(void *cmdList) {
  char *cmdptr = (char *)cmdList;
  register int tok, cmdsize;

  MSGDEBUG(2, "Default Display Renderer.");
  MSGDEBUG(2, sendmsg);

  if(!cmdList)
    return;
  
  // save transformation matrix
  transMat.push();

  // scan through the list, getting each command and executing it, until
  // the end of commands token is found
  while((tok = ((int *)cmdptr)[0]) != DLASTCOMMAND) {
    if (tok == DLINKLIST) {
       cmdptr += sizeof(int);       // step forward one (no length here)
       cmdptr = *((char **) cmdptr);// and go to the next link
       tok = ((int *)cmdptr)[0];    // this is guarenteed to be neither
    }                               // a DLINKLIST nor DLASTCOMMAND
       
    cmdsize = ((int *)cmdptr)[1];
    cmdptr += 2*sizeof(int);

    MSGDEBUG(4, "DisplayDevice: doing command " << tok);
    MSGDEBUG(4, " of size " << cmdsize << sendmsg);

    if(tok == DCOLORINDEX) {
      // set the current color to the given color index ... assumes the
      // color has already been defined
      int n = ((int *)cmdptr)[0];
      
      if(!matDefined[n]) {
        msgErr << "DisplayDevice: Error, color " << n << " not defined."
	       << sendmsg;
      } else {
        // bind the material if materials enabled
	materialOn = n;
	if(materialsActive)
	  activate_material();
      }

    } else if(tok == DCOLORRGB) {
      // just set the non-material color to the given value.  Turns off
      // materials.
      deactivate_materials();

    } else if(tok == DLIGHTONOFF) {
      // turn on or off a light source
      activate_light(((int *)cmdptr)[0], ((int *)cmdptr)[1]);
      
    } else if(tok == DMATERIALS) {
      // enable/disable material characteristics
      if(((int *)cmdptr)[0])
        activate_material();
      else
        deactivate_materials();
      
    } else if(tok == DSPHERERES) {
      // set the current sphere resolution
      sphereRes = ((int *)cmdptr)[0];
      
    } else if(tok == DSPHERETYPE) {
      // set the current sphere type
      sphereMode = ((int *)cmdptr)[0];
      
    } else if(tok == DLINESTYLE) {
      // set the current line style
      lineStyle = ((int *)cmdptr)[0];
      
    } else if(tok == DLINEWIDTH) {
      // set the current line width
      lineWidth = ((int *)cmdptr)[0];

    } else if(tok == DCOLORDEF) {
      // define a new color
      int n = (int)(((float *)cmdptr)[0]);
      define_material(n, ((float *)cmdptr) + 1);
      
    } else if(tok == DLIGHTDEF) {
      // define a new light source
      int n = (int)(((float *)cmdptr)[0]);
      define_light(n, ((float *)cmdptr) + 1, ((float *)cmdptr) + 4);
      
    } else if(tok == DCLEAR) {
      // clear the display
      clear();

    } else if(tok == DPUSH) {
      // push the current trans matrix
      transMat.push();

    } else if(tok == DPOP) {
      // pop the current trans matrix
      transMat.pop();
      
    } else if(tok == DLOAD) {
      // load the given trans matrix onto the stack
      Matrix4 tmpmat((float *)cmdptr);
      (transMat.top()).loadmatrix(tmpmat);

    } else if(tok == DMULT) {
      // multiply the given trans matrix
      Matrix4 tmpmat((float *)cmdptr);
      (transMat.top()).multmatrix(tmpmat);

    } else if(tok == DTRANS) {
      // do a translation
      (transMat.top()).translate(((float *)cmdptr)[0], ((float *)cmdptr)[1],
      		((float *)cmdptr)[2]);
      
    } else if(tok == DSCALE) {
      // do a scale
      (transMat.top()).scale(((float *)cmdptr)[0], ((float *)cmdptr)[1],
      		((float *)cmdptr)[2]);

    } else if(tok == DROT) {
      // do a rotation about a given axis
      (transMat.top()).rot(((float *)cmdptr)[0],
      		'x' + (int)(((float *)cmdptr)[1]));

    } else {
      // command not found, just ignore it here
    }
    
    // update position in array
    cmdptr += cmdsize;
  }
  
  // restore transformation matrix
  transMat.pop();
}

// Define render_done.  This could probably be moved to .h
void DisplayDevice::render_done() {}


//
//*******************  the picking routine  *********************
//
// This scans the given command list until the end, finding which item is
// closest to the given pointer position.
//
// arguments are dimension of picking (2 or 3), position of pointer,
// draw command list, and returned distance from object to eye position.
// Returns ID code ('tag') for item closest to pointer, or (-1) if no pick.
// If an object is picked, the eye distance argument is set to the distance
// from the display's eye position to the object (after its position has been
// found from the transformation matrix).  If the value of the argument when
// 'pick' is called is <= 0, a pick will be generated if any item is near the
// pointer.  If the value of the argument is > 0, a pick will be generated
// only if an item is closer to the eye position than the value of the
// argument.
// For 2D picking, coordinates are relative position in window from
//	lower-left corner (both in range 0 ... 1)
// For 3D picking, coordinates are the world coords of the pointer.  They
//	are the coords of the pointer after its transformation matrix has been
//	applied, and these coordinates are compared to the coords of the objects
//	when their transformation matrices are applied.

// but first, a macro for returning the distance^2 from the eyepos to the
// given position
#define DTOEYE(x,y,z) ( (x-eyePos[0])*(x-eyePos[0]) + \
			(y-eyePos[1])*(y-eyePos[1]) + \
			(z-eyePos[2])*(z-eyePos[2]) )

int DisplayDevice::pick(int dim, float *pos, void *cmdList, float &eyedist) {
  char *cmdptr = (char *)cmdList;
  int tok, cmdsize;
  float newEyeDist, currEyeDist = eyedist;
  int tag = (-1), inRegion, currTag, minX, minY, maxX, maxY;
  float fminX, fminY, fminZ, fmaxX, fmaxY, fmaxZ;
  float pntpos[3];
  long cpos[2];
#ifdef VMDGL
  // GL doesn't set the cpos if the 3d point is outside the screen position.
  // There is no way to tell if this occurs, so as far as the routine can
  // tell, the point off the screen has the same screen coord as the previous
  // point which was on.  What I do is save the last cpos.  If the new one
  // has the same coordinate as the old, I throw it away.  This will fail
  // if: 1) atoms i, j are on the same point and all of (i+1 .. j-1) are
  // off the screen.  2) The first atoms are off the screen and the current
  // mouse position is near the previous pick point.
  long last_cpos[2];
  last_cpos[0] = -1;
  last_cpos[1] = -1;
#endif

  MSGDEBUG(3, "DisplayDevice: checking for a picked object." << sendmsg);
  
  if(!cmdList)
    return (-1);
  
  // initialize picking: find screen region to look for object
  if(dim == 2) {
    fminX = pos[0] - pick_region();
    fmaxX = pos[0] + pick_region();
    fminY = pos[1] - pick_region();
    fmaxY = pos[1] + pick_region();
    abs_screen_pos(fminX, fminY);
    abs_screen_pos(fmaxX, fmaxY);
    minX = (int)fminX;
    maxX = (int)fmaxX;
    minY = (int)fminY;
    maxY = (int)fmaxY;
  } else {
    fminX = pos[0] - 5.0 * pick_region();
    fmaxX = pos[0] + 5.0 * pick_region();
    fminY = pos[1] - 5.0 * pick_region();
    fmaxY = pos[1] + 5.0 * pick_region();
    fminZ = pos[2] - 5.0 * pick_region();
    fmaxZ = pos[2] + 5.0 * pick_region();
  }

  // make sure we do not disturb the regular transformation matrix
  transMat.push();

  // scan through the list, getting each command and executing it, until
  // the end of commands token is found
  float *dataBlock = NULL;
  while((tok = ((int *)cmdptr)[0]) != DLASTCOMMAND) {
    if (tok == DLINKLIST) {
       cmdptr += sizeof(int);       // step forward one (no length here)
       cmdptr = *((char **) cmdptr);// and go to the next link
       tok = ((int *)cmdptr)[0];    // this is guarenteed to be neither
    }                               // a DLINKLIST nor DLASTCOMMAND
       
    cmdsize = ((int *)cmdptr)[1];
    cmdptr += 2*sizeof(int);

    MSGDEBUG(4, "DisplayDevice: doing pick command " << tok);
    MSGDEBUG(4, " of size " << cmdsize << sendmsg);

    if(tok == DPICKPOINT || tok == DPICKPOINT_I) {
    
      // calculate the transformed position of the point
      if(tok == DPICKPOINT) {
        (transMat.top()).multpoint3d((float *)cmdptr, pntpos);
	currTag = (int) ( ((float *)cmdptr)[3]);
      } else {
        (transMat.top()).multpoint3d(dataBlock + ((int *)cmdptr)[0], pntpos);
 	currTag = ((int *)cmdptr)[1];
      }

      // check if in picking region ... different for 2D and 3D
      if(dim == 2) {
        // convert the 3D world coordinate position to 2D absolute screen coord
	abs_screen_loc_3D(pntpos, cpos);
      
        // check to see if the position falls in our picking region
        inRegion = (cpos[0] >= minX && cpos[0] <= maxX &&
			cpos[1] >= minY && cpos[1] <= maxY);
#ifdef VMDGL
      // check that this isn't the same as the previous point
      if (cpos[0] == last_cpos[0] && cpos[1] == last_cpos[1]) {
        inRegion = FALSE;
      }
      last_cpos[0] = cpos[0];
      last_cpos[1] = cpos[1];
#endif

      } else {
        // just check to see if the position is in a box centered on our
	// pointer.  The pointer position should already have been transformed.
	inRegion = (pntpos[0] >= fminX && pntpos[0] <= fmaxX &&
			pntpos[1] >= fminY && pntpos[1] <= fmaxY &&
			pntpos[2] >= fminZ && pntpos[2] <= fmaxZ);
      }
      
      // has a hit occurred?
      if(inRegion) {
        // yes, see if it is closer to the eye position than earlier objects
	newEyeDist = DTOEYE(pntpos[0], pntpos[1], pntpos[2]);
	if(currEyeDist < 0.0 || newEyeDist < currEyeDist) {
	  currEyeDist = newEyeDist;
	  tag = currTag;
	}
      }

    } else if(tok == DPICKLINE || tok == DPICKLINE_I
		|| tok == DPICKBOX || tok == DPICKBOX_I) {
      ;		// do nothing, these are as-yet unsupported picking tokens

    } else if(tok == DPUSH) {
      // push the current trans matrix
      transMat.push();

    } else if(tok == DPOP) {
      // pop the current trans matrix
      transMat.pop();
      
    } else if(tok == DLOAD) {
      // load the given trans matrix onto the stack
      Matrix4 tmpmat((float *)cmdptr);
      (transMat.top()).loadmatrix(tmpmat);

    } else if(tok == DMULT) {
      // multiply the given trans matrix
      Matrix4 tmpmat((float *)cmdptr);
      (transMat.top()).multmatrix(tmpmat);

    } else if(tok == DTRANS) {
      // do a translation
      (transMat.top()).translate(((float *)cmdptr)[0], ((float *)cmdptr)[1],
      		((float *)cmdptr)[2]);
      
    } else if(tok == DSCALE) {
      // do a scale
      (transMat.top()).scale(((float *)cmdptr)[0], ((float *)cmdptr)[1],
      		((float *)cmdptr)[2]);

    } else if(tok == DROT) {
      // do a rotation about a given axis
      (transMat.top()).rot(((float *)cmdptr)[0],
      		'x' + (int)(((float *)cmdptr)[1]));

    } else if(tok == DMOREDATA || tok == DNEWDATA) {
      // set the current drawing data block
#ifdef VMDCAVE
      dataBlock = (float *)cmdptr;      // point to current position in list
#else
      dataBlock = ((float **)cmdptr)[0];        // point to given memory loc
#endif

    }
    
    // update position in array
    cmdptr += cmdsize;
  }

  // make sure we do not disturb the regular transformation matrix
  transMat.pop();

  // return result; if tag >= 0, we found something
  eyedist = currEyeDist;
  return tag;
}

