/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: DrawMolItem.h,v $
 *	$Author: ulrich $	$Locker:  $		$State: Exp $
 *	$Revision: 1.27 $	$Date: 1997/03/21 22:17:42 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * Child Displayable component of a molecule; this is responsible for doing
 * the actual drawing of a molecule.  It contains an atom color, atom
 * selection, and atom representation object to specify how this component
 * should look.
 *
 ***************************************************************************/
#ifndef DRAWMOLITEM_H
#define DRAWMOLITEM_H

#include "Displayable.h"
#include "DispCmds.h"
#include "AtomColor.h"
#include "AtomRep.h"
#include "AtomSel.h"
#include "ColorList.h"

// forward declaration
class DrawMolecule;


class DrawMolItem : public Displayable3D {

public:
  // atom coloring method to use
  AtomColor *atomColor;
  
  // atom representation to use
  AtomRep *atomRep;
  
  // atom selection to use
  AtomSel *atomSel;

  // current representation ID
  int repNumber;

  // buffer for comments  
  char commentBuffer[1024];

  // used so AtomRep can get the number of dot surfaces to use
  // declared and defined in DrawMolItemSurfacePoints.data
  static int num_dot_surfaces;

private:
  // parent molecule; kept here as a 'DrawMolecule' pointer for convenience
  DrawMolecule *mol;
  
  // useful drawing command objects, used to create display list
  DispCmdData *dataCmd;		// put in block of data with coordinates
  DispCmdColorIndex cmdColorIndex;
  DispCmdColorRGB cmdColorRGB;
  DispCmdTextPosIndex cmdTextPosIndex;
  DispCmdSphereIndex cmdSphereIndex;
  DispCmdLine cmdLine;
  DispCmdTriangle cmdTriangle;
  DispCmdCylinder cmdCylinder;
  DispCmdText cmdTextX;
  DispCmdComment cmdCommentX;
  DispCmdMaterials cmdMaterials;
  DispCmdSphereRes cmdSphres;
  DispCmdSphereType cmdSphtype;
  DispCmdLineType cmdLineType;
  DispCmdLineWidth cmdLineWidth;
  
  // put in commands to draw with different representations, using the
  // given positions
  void draw_just_spheres(void);	// used for vdw and dotted reps
  void draw_lines(float *);
  void draw_cpk(float *);
  void draw_points(float *);
  void draw_vdw(void);
  void draw_dotted(void);
  void draw_licorice(float *);
  void draw_bonds(float *);
  void make_connection(float *prev, float *start, float *end, float *next,
		       float radius, int resolution, int is_cyl);
  void draw_spline_curve(int num, float *coords, int *idx,
			 int use_cyl, float b_rad, int b_res);
  void draw_spline_ribbon(int num, float *coords, float *perps,
			  int *idx, int use_cyl, float b_rad,
			  int b_res);
  void draw_tube(float *);
  void draw_ribbons(float *);  // use an external call to 'ribbons'
  void draw_structure(float *);  // based on structure
  void draw_alpha_helix_cylinders(ResizeArray<float> &x,
				  ResizeArray<float> &y,
				  ResizeArray<float> &z,
				  ResizeArray<int> &atom_on, int *color,
				  float bond_rad, int bond_res,
				  float *res_start, float *res_end);
  void draw_beta_sheet(ResizeArray<float> &x, ResizeArray<float> &y,
		       ResizeArray<float> &z, ResizeArray<int> &atom_on, 
		       int *color, float ribbon_width,
		       float *res_start, float *res_end);
  void draw_trace(float *pos); // C-alpha trace
  void draw_dot_surface(float *pos); // dot surface, kinda
#ifdef VMDALPHA_SHAPE
  void draw_alphaShape(float *); // draw the alpha shape.  Only available
                       // in the Theoretical Biophysics Group as it
		       // is still under development
#endif
  void draw_surface(float *pos);
  void draw_hbonds(float *);

  // regenerate the command list
  void create_cmdlist(void);

  // do we need to completely redraw the list?
  // This keeps track of what caused the regeneration
public:
  enum RegenChoices {NO_REGEN = 0, FORCE_REGEN = 1, SEL_REGEN = 2,
	REP_REGEN = 4, COL_REGEN = 8};
private:
  int needRegenerate;

  // do we need to update the current frame coordinates?
  int needNewFrame;

  // add the pick points
  void place_picks(void);

protected:
  // do action due to the fact that a color for the given ColorList for
  // the specified category has changed
  virtual void do_color_changed(ColorList *, int);

public:
  // constructor: parent molecule, and atom drawing methods
  DrawMolItem(DrawMolecule *, AtomColor *, AtomRep *, AtomSel *);
  virtual ~DrawMolItem(void);
  
  // change the atom coloring method.  Return success.
  int change_color(AtomColor *);
  
  // change the atom representation method.  Return success.
  int change_rep(AtomRep *);
  
  // change the atom selection method.  Return success.
  int change_sel(AtomSel *);

  // force a recalculation of everything
  void force_recalc(RegenChoices  = FORCE_REGEN);

  // return whether the Nth atom is displayed.  This is true if ANY
  // representation is displaying the given atom
  int atom_displayed(int);

  // return which representation this is
  int representation_index(void);

  //
  // public virtual routines
  //

  // Create a popup menu, with the given name; return NULL if none available
  // subclasses should have the parent first create a menu, and then add items
  // to the end of it.  By default, this does nothing and returns NULL.
  // 1st argument is the tag of the item selected (it should correspond to
  // a tag from this object).
  virtual PopupMenu *create_popup_menu(int);

  // prepare for drawing ... do any updates needed right before draw.
  virtual void prepare(DisplayDevice *);

};

#endif

